# Econ 70310: Numerical Methods in Economics
# Fall 2017
# Lecture 1.2: Basics
# Example Code
# Zach Stangebye, Aug 2017

# If you have a bunch of variables you want to clear to start a new project,
# you can do so with the workspace() command

# BE AWARE: To run this full code on your own machine, you'll need to add three
# packages first by typing the following into the console
# Pkg.add("Gadfly") # the plotting package
# Pkg.add("Cairo") # necessary for the PDF backend of the plotter
# Pkg.add("Fontconfig") # necessary for the PDF backend of the plotter


x = 1.0
workspace()
# don't use workspace prior to plotting. It ruins the Juno plotting functionality.
# Exports to pdf/png still work fine, though
# Avoid using it, though...it can also ruin other packages (such as JuMP or NLopt)

# variables initialized in the main body of the code will remain in your
# workspace as global variables
x = 10.0
y = 1.0

# variables defined in loops or functions will not
for i in 1:3
  z = 1
  z = z + i
end
# none of these variables will appear in the workspace

# We can change this with the local and global prefixes
for i in 1:3
  global z = 1
  z = z + i
end
# Now z will appear in the workspace

# We can also tell a substructure to consider ignore global variables of the
# same name
for i in 1:3
  x = x + i
end
# This loop alters the value of the global variable x

for i in 1:3
  local x = 0
  x = x + i
  if i == 3
    println(x)
  end
end
# This loop leaves the global variable x untouched, operating on a new
# local variable with the same name

# Scope matters for time evaluation. Global variables slower
b = 1.0
function g(a)
  for i in 1:1000000
    tmp = a + b
  end
end

# making b local boosts speed considerably
function g(a, b)
  for i in 1:1000000
    tmp = a + b
  end
end
# Julia will reference latter function if g() has two arguments; the former
# if it has one

@time g(1.0)
@time g(1.0,1.0)

# An alternative way to improve speed is to make global variables parameters
# instead of variables. Can do this with the

const η = 1.0
function f(a)
  for i in 1:1000000
    tmp = η + a
  end
end

@time f(1.0)

# It works either way, but local variables orders of magnitude faster. A good way to
# fine-tune your code and make it faster is to make more stuff local






# Maximization approach 1: Grid-search

function grid_search(N)
  p = .3 # specify a price
  # endowments
  e_11 = 4.6
  e_12 = .23

  # Form a grid of N points
  c_lb = 0.001
  c_ub = (p*e_11 + (1-p)*e_12)/p - .001 # give some space at the bounds to avoid -Inf

  # Make a grid with the linspace command:
  c_grid_ls = linspace(c_lb,c_ub,N)

  # Another alternative is to define a Floatrange with colon
  c_grid_fr = c_lb:(c_ub-c_lb)/(N-1):c_ub

  # linspace and floatrange are not quite arrays. Less memory, but Less
  # functionality. Can turn them into arrays easily with collect
  c_grid = collect(c_grid_fr)

  # preferences
  β_11 = 2.2 # making them parameters makes evaluations later move more quickly
  β_12 = 4.5

  # We now define our objective function
  utility(c) = β_11*log.(c) + β_12*log.( (p*(e_11-c) + (1-p)*e_12)/(1-p) )

  # with JIT compilation, we can give it an array in this case, since log() takes
  # array arguments
  utility_vec = utility(c_grid)
  # More generally, we can apply a . after the function e.g. utility.(c_grid) to evaluate
  # it over an array, even if the function itself can only take scalars

  # We now pick the largest value
  u_star = maximum(utility_vec) # this tells us the highest utility
  soln_tuple = findmax(utility_vec) # this tells us the highest utility and the gridpoint

  c_star = c_grid[soln_tuple[2]]

  # Compare with true solution
  c_star_true = β_11/(β_11 + β_12)*(p*e_11 + (1-p)*e_12)/p

  soln_error = abs(c_star_true - c_star)
  utility_error = abs(u_star - utility(c_star_true))
  return [c_star, soln_error, utility_error]
end
# Can reduce errors by increasing N: Requires more time
# Function evaluations scale by N^(dimension of state space)

# We can check the time by putting it into a function
@time grid_search(10)
@time grid_search(10) # 2nd time always faster: JIT compilation
@time grid_search(1000)
@time grid_search(1000000) # tradeoff between time and accuracy

function grid_search_modified(N)
  p = .3 # specify a price

  # endowments
  e_11 = 4.6
  e_12 = .23

  # Form a grid of N points
  c_lb = 0.001
  c_ub = (p*e_11 + (1-p)*e_12)/p - .001 # give some space at the bounds to avoid -Inf

  # Make a grid with the linspace command:
  c_grid_ls = linspace(c_lb,c_ub,N)

  # Another alternative is to define a Floatrange with colon
  c_grid_fr = c_lb:(c_ub-c_lb)/(N-1):c_ub

  # linspace and floatrange are not quite arrays. Less memory, but Less
  # functionality. Can turn them into arrays easily with collect
  c_grid = collect(c_grid_fr)

  # preferences
  β_11 = 2.2 # making them parameters makes evaluations later move more quickly
  β_12 = 4.5

  # We now define our objective function
  utility(c) = β_11*log.(c) + β_12*log.( (p*(e_11-c) + (1-p)*e_12)/(1-p) )

  # with JIT compilation, we can give it an array in this case, since log() takes
  # array arguments
  utility_vec = utility(c_grid)
  # More generally, we can apply a . after the function e.g. utility.(c_grid) to evaluate
  # it over an array, even if the function itself can only take scalars

  # We now pick the largest value
  u_star = utility(c_grid[1])
  c_star = c_grid[1]
  number_evals = 1
  for ic in 2:N
    # look for a turn-around
    current_u = utility(c_grid[ic])
    if current_u <= u_star
      number_evals = ic
      break
    else
      u_star = current_u
    end
  end

  # Compare with true solution
  c_star_true = β_11/(β_11 + β_12)*(p*e_11 + (1-p)*e_12)/p

  soln_error = abs(c_star_true - c_star)
  utility_error = abs(u_star - utility(c_star_true))
  return [c_star, soln_error, utility_error, number_evals/N]
end

@time grid_search_modified(10)
@time grid_search_modified(10) # 2nd time always faster: JIT compilation
@time grid_search_modified(1000)
@time grid_search_modified(1000000)

# Can see that it takes a lot fewer function evaluations: sometimes can approach the
# built-in machine-level functionality of the findmax() function

function golden_search(opt_tol,pprice,inc_factor)
  # opt_tol is the tolerance parameter

  # endowments
  e_11 = 4.6
  e_12 = .23

  c_lb = 0.001
  c_ub = (pprice*e_11 + (1-pprice)*e_12)/pprice - .001

  α_1 = (3-√(5))/2.0
  α_2 = (√(5)-1)/2.0

  # preferences
  β_11 = 2.2 # making them parameters makes evaluations later move more quickly
  β_12 = 4.5

  # endowments
  e_11 = 4.6*inc_factor
  e_12 = .23*inc_factor

  utility(c) = β_11*log.(c) + β_12*log.( (pprice*(e_11-c) + (1-pprice)*e_12)/(1-pprice) )

  c1 = c_lb+ α_1*(c_ub-c_lb)
  c2 = c_lb + α_2*(c_ub-c_lb)
  v1 = utility(c1)
  v2 = utility(c2)
  d = α_1*α_2*(c_ub-c_lb)
  ccounter = 2
  while d > opt_tol
    ccounter += 1
    d = d*α_2
    if v2 < v1
      c2 = c1
      c1 = c1-d
      v2 = v1
      v1 = utility(c1)
    else
      c1 = c2
      c2 = c2+d
      v1 = copy(v2)
      v2 = utility(c2)
    end
  end
  if v2 > v1
    c_star = c2
    u_star = v2
  else
    c_star = c1
    u_star = v1
  end
  ccounter += 1
  # Compare with true solution
  c_star_true = β_11/(β_11 + β_12)*(pprice*e_11 + (1-pprice)*e_12)/pprice

  soln_error = abs(c_star_true - c_star)
  utility_error = abs(u_star - utility(c_star_true))
  return [c_star, soln_error, utility_error, ccounter]
end


# Gets us pretty good precision in less than 30 evaluations
# grid-search requires tens of thousands to millions to get similar tolerance
# self-coded, but faster and more accurate than black-box grid-search

@time golden_search(1e-4,.3,1.0)


# Now, we use our technique to build demand curves.
# First, we'll build a price grid search across prices
price_grid = 0.01:0.01:0.99
ttol = 1e-4
demand_curve = similar(collect(price_grid)) # pre-allocate a demand curve with the `similar` function

income_factor = 1.0
for ip in 1:size(price_grid,1) # can add @time before any block of code (including for block)
  ttemp = golden_search(ttol,price_grid[ip],income_factor)
  demand_curve[ip] = ttemp[1]
end


using Gadfly # This is (one of) the plotting package(s)
# Before this command, will have to say enter: Pkg.add("Gadfly") into console
# Same for any other package

# don't use workspace() command before you plot...for some reason this throws off Gadfly

# by default, it plots in the juno plot pane; however, must execute separately.
# plot commands are skipped over in a full-file execution
plot(x = price_grid, y = demand_curve, Geom.line,
  Guide.xlabel("Price"), Guide.ylabel("Qty Demanded"), Guide.title("Demand Curve") )

# Now let's use it for analysis. One quick thing we can do is see how it responds
# to income shifts
demand_curve2 = similar(collect(price_grid)) # pre-allocate a demand curve with the `similar` function

income_factor = 2.0
for ip in 1:size(price_grid,1) # can add @time before any block of code (including for block)
  ttemp = golden_search(ttol,price_grid[ip],income_factor)
  demand_curve2[ip] = ttemp[1]
end

p = plot(layer(x = price_grid, y = demand_curve, Geom.line,Theme(default_color=color("orange"))),
  layer(x = price_grid, y = demand_curve2, Geom.line,Theme(default_color=color("purple"))),
  Guide.xlabel("Price"), Guide.ylabel("Qty Demanded"), Guide.title("Demand Curve") )

# in addition, can send plots to a clean pdf or png file. These will execute in-line,
# unlike Juno plots
draw(PDF("sample.pdf",7inch,5inch),p)

# Newton-Raphson Method to solve our problem

function newton_raphson(opt_tol,pprice,c_star)
  # opt_tol is the tolerance parameter; c_star is the initial guess

  # preferences
  β_11 = 2.2 # making them parameters makes evaluations later move more quickly
  β_12 = 4.5

  # endowments
  e_11 = 4.6
  e_12 = .23

  c_lb = 0.0
  c_ub = (pprice*e_11 + (1-pprice)*e_12)/pprice

  utility(c) = β_11*log.(c) + β_12*log.( (pprice*(e_11-c) + (1-pprice)*e_12)/(1-pprice) )
  uPrime(c) = β_11./c - β_12*(pprice/(1-pprice))./( (pprice*(e_11-c) + (1-pprice)*e_12)/(1-pprice) )
  uDoublePrime(c) = -β_11./c.^2 - β_12*(pprice/(1-pprice)).^2./( (pprice*(e_11-c) + (1-pprice)*e_12)/(1-pprice) ).^2

  # start low (starting at /2.0 for p=4 only takes 2 iterations)
  grad_current = uPrime(c_star)
  ccounter = 1
  while ( abs( grad_current ) > opt_tol) & (ccounter < 500)
    ccounter += 1
    hess_current = uDoublePrime(c_star)
    c_star = c_star - grad_current/hess_current
    grad_current = uPrime(c_star)
  end

  u_star = utility(c_star)
  # Compare with true solution
  c_star_true = β_11/(β_11 + β_12)*(pprice*e_11 + (1-pprice)*e_12)/pprice

  soln_error = abs(c_star_true - c_star)
  utility_error = abs(u_star - utility(c_star_true))
  return [c_star, soln_error, utility_error, ccounter]
end

# Newton-Raphson with numerical derivatives
function newton_raphson_numerical(opt_tol,pprice,c_star)
  # opt_tol is the tolerance parameter; c_star is the initial guess

  # preferences
  β_11 = 2.2 # making them parameters makes evaluations later move more quickly
  β_12 = 4.5

  # endowments
  e_11 = 4.6
  e_12 = .23

  c_lb = 0.0
  c_ub = (pprice*e_11 + (1-pprice)*e_12)/pprice

  utility(c) = β_11*log.(c) + β_12*log.( (pprice*(e_11-c) + (1-pprice)*e_12)/(1-pprice) )

  step_size = 1e-3
  # start low (starting at /2.0 for p=4 only takes 2 iterations)
  grad_current = (utility(c_star+step_size) - utility(c_star-step_size))/(2*step_size)
  ccounter = 1
  while ( abs( grad_current ) > opt_tol) & (ccounter < 500)
    ccounter += 1
    hess_current = (utility(c_star+step_size) + utility(c_star-step_size) - 2*utility(c_star))/(step_size^2)
    c_star = c_star - grad_current/hess_current
    grad_current = (utility(c_star+step_size) - utility(c_star-step_size))/(2*step_size)
  end

  u_star = utility(c_star)
  # Compare with true solution
  c_star_true = β_11/(β_11 + β_12)*(pprice*e_11 + (1-pprice)*e_12)/pprice

  soln_error = abs(c_star_true - c_star)
  utility_error = abs(u_star - utility(c_star_true))
  return [c_star, soln_error, utility_error, ccounter]
end


# Now we'll try it with Julia's autodifferentiation
using ForwardDiff

function newton_raphson_auto(opt_tol,pprice,c_star)
  # opt_tol is the tolerance parameter

  # preferences
  β_11 = 2.2 # making them parameters makes evaluations later move more quickly
  β_12 = 4.5

  # endowments
  e_11 = 4.6
  e_12 = .23

  c_lb = 0.0
  c_ub = (pprice*e_11 + (1-pprice)*e_12)/pprice

  utility(c) = β_11*log.(c) + β_12*log.( (pprice*(e_11-c) + (1-pprice)*e_12)/(1-pprice) )
  uPrime(c) = ForwardDiff.derivative(utility,c)
  uDoublePrime(c) = ForwardDiff.derivative(uPrime,c)

  grad_current = uPrime(c_star)
  ccounter = 1
  while ( abs( grad_current ) > opt_tol) & (ccounter < 500)
    ccounter += 1
    hess_current = uDoublePrime(c_star)
    c_star = c_star - grad_current/hess_current
    grad_current = uPrime(c_star)
  end

  u_star = utility(c_star)
  # Compare with true solution
  c_star_true = β_11/(β_11 + β_12)*(pprice*e_11 + (1-pprice)*e_12)/pprice

  soln_error = abs(c_star_true - c_star)
  utility_error = abs(u_star - utility(c_star_true))
  return [c_star, soln_error, utility_error, ccounter]
end

# Another very useful package is Julia's JLD package, which can be used to save and load
# variables

using JLD
xvar = 1.0
yarr = [1.0, 2.0]

@save "test_save.jld" xvar yarr

# shut Julia down. Reopen, enter

# using JLD
# @load "test_save.jld"

# and xvar and yarr will be part of the workspace again
