# Econ 70310: Numerical Methods in Economics
# Fall 2017
# Lecture 1.3: Basics
# Example Code
# Zach Stangebye, Aug 2017

# Built-in optimization packages

# If you don't want to use your own algorithms, there are a couple of options

using Optim
# The Optim package handles univariate and multivariate optimization problems,
# but can only do simple upper and lower bounds. Advanced constraints are too much
# for it

function optimExample(pprice)
  # opt_tol is the tolerance parameter; c_star is the initial guess

  # preferences
  β_11 = 2.2 # making them parameters makes evaluations later move more quickly
  β_12 = 4.5

  # endowments
  e_11 = 4.6
  e_12 = .23

  c_lb = 0.0
  c_ub = (pprice*e_11 + (1-pprice)*e_12)/pprice

  utility(c) = β_11*log.(c) + β_12*log.( (pprice*(e_11-c) + (1-pprice)*e_12)/(1-pprice) )

  negUtility(c) = -utility(c) # by default, Optim minimizes a function
  # pass it lower and upper bounds
  result = Optim.optimize(negUtility,c_lb,c_ub)

  c_star = result.minimizer
  u_star = -result.minimum
  # Compare with true solution
  c_star_true = β_11/(β_11 + β_12)*(pprice*e_11 + (1-pprice)*e_12)/pprice

  soln_error = abs(c_star_true - c_star)
  utility_error = abs(u_star - utility(c_star_true))
  return [c_star, soln_error, utility_error]
end

# Optim defaults to Brent's method for single-variable problem (sort of a hybrid
# of a bunch of stuff we've learned). We can tell it to use a different algorithm
# if we want. In this example, you can tell it to do a golden-search as an option
function optimExampleGoldenSearch(pprice)
  # opt_tol is the tolerance parameter; c_star is the initial guess

  # preferences
  β_11 = 2.2 # making them parameters makes evaluations later move more quickly
  β_12 = 4.5

  # endowments
  e_11 = 4.6
  e_12 = .23

  c_lb = 0.0
  c_ub = (pprice*e_11 + (1-pprice)*e_12)/pprice

  utility(c) = β_11*log.(c) + β_12*log.( (pprice*(e_11-c) + (1-pprice)*e_12)/(1-pprice) )

  negUtility(c) = -utility(c) # by default, Optim minimizes a function
  # pass it lower and upper bounds
  result = Optim.optimize(negUtility,c_lb,c_ub,GoldenSection())

  c_star = result.minimizer
  u_star = -result.minimum
  # Compare with true solution
  c_star_true = β_11/(β_11 + β_12)*(pprice*e_11 + (1-pprice)*e_12)/pprice

  soln_error = abs(c_star_true - c_star)
  utility_error = abs(u_star - utility(c_star_true))
  return [c_star, soln_error, utility_error]
end


using JuMP # this is the convex programming package
using NLopt # this is the solver the package will use
# for pure linear programming, the Cbc package has a nice solver as well

# utility maximization example: Using JuMP for constrained, multivariate
# maximization
function NLOptExampleGradient(price_apples,price_oranges,income)

  # first, set the parameters
  β_11 = .3
  β_12 = .7

  # price_apples = 1.1
  # price_oranges = 3.1
  # income = 10

  # Define a constrained maximization program
  # Use the Method-of-Moving-Asymptotes (MMA): Gradient-based search for local max
  demand_prob = Model(solver=NLoptSolver(algorithm=:LD_MMA))

  # add the choice variables
  @variable(demand_prob,apples >= 0)
  @variable(demand_prob,oranges >= 0)

  # Set the initial values for the choice variables
  setvalue(apples,.5*income/price_apples)
  setvalue(oranges,.5*income/price_oranges)

  # add the budget constraint (and any other constraints)
  @constraint(demand_prob,apples*price_apples+oranges*price_oranges <= income)

  # set the objective function (and tell it it's a max instead of a min)
  @NLobjective(demand_prob,Max,β_11*log(apples)+β_12*log(oranges))

  # solve the problem!
  solve(demand_prob)

  # To get the results...
  println("Maximum Utility: ", getObjectiveValue(demand_prob))
  println("Demand")
  println(getValue(apples)," apples")
  println(getValue(oranges)," oranges")

  return [getObjectiveValue(demand_prob),getValue(apples),getValue(oranges)]
end

# Cautionary note: JuMP for gradient-based functions makes use of Julia's autodifferentiation. If you give it
# an objective that's not compatible with that e.g. a function with loops,
# it will give you an error and break down


function NLOptExamplePatternSearch(price_apples,price_oranges,income)

  # first, set the parameters
  β_11 = .3
  β_12 = .7

  # price_apples = 1.1
  # price_oranges = 3.1
  # income = 10

  # Define a constrained maximization program
  # Use the Nelder-Mead (pattern-search): Gradient-free search for global
  demand_prob = Model(solver=NLoptSolver(algorithm=:LN_COBYLA))

  # add the choice variables
  @variable(demand_prob,apples >= 0)
  @variable(demand_prob,oranges >= 0)

  # Set the initial values for the choice variables
  setvalue(apples,.5*income/price_apples)
  setvalue(oranges,.5*income/price_oranges)

  # add the budget constraint (and any other constraints)
  @constraint(demand_prob,apples*price_apples+oranges*price_oranges <= income)

  # set the objective function (and tell it it's a max instead of a min)
  @NLobjective(demand_prob,Max,β_11*log(apples)+β_12*log(oranges))

  # solve the problem!
  solve(demand_prob)

  # To get the results...
  println("Maximum Utility: ", getObjectiveValue(demand_prob))
  println("Demand")
  println(getValue(apples)," apples")
  println(getValue(oranges)," oranges")

  return [getObjectiveValue(demand_prob),getValue(apples),getValue(oranges)]
end
