# Econ 70310: Numerical Methods in Economics
# Fall 2017
# Lecture 1.4: Basics
# Example Code
# Zach Stangebye, Aug 2017

# Numerical expectations and integration

# To access a wide array of useful distribution information, use the distributions
# package

using Distributions
using Gadfly
# Define a random-variable object

# We'll try it first for a normal
function normal_test!(garbage)
  # First, set the relevant parameters
  mean_x = 2.3
  std_x = 4.65

  # Next, define the variable
  x = Normal(mean_x,std_x)

  # You can ask it for the pdf

  y = pdf(x,8.4)

  # You could even evaluate it at a bunch of points and plot it
  N = 1000
  ggrid = mean_x-3std_x:6std_x/(N-1): mean_x+3std_x

  pdf_vec = pdf(x,ggrid)

  plot(x=ggrid,y=pdf_vec,Geom.line)

  # You can do the same for the CDF
  cdf_vec = cdf(x,ggrid)

  plot(x=ggrid,y=cdf_vec,Geom.line)

  # you can get the relevant parameters with the params() command
  x_params = params(x)

  # You can also compute a bunch of moments
  x_mean = mean(x)
  x_median = median(x)
  x_mode = mode(x) # returns first mode if there are multiple
  x_skewness = skewness(x)
  x_kurtosis = kurtosis(x) # excess kurtosis over normal
  x_5perc = quantile(x,.05)
  x_95perc = quantile(x,.95)

  # You can even get the entropy!
  x_entropy = entropy(x)

  # And the log-likelihood of a bunch of draws
  x_sample = [2.3,7.4,2.3]
  sampleLL = loglikelihood(x,x_sample)

  # You can also take random draws from it
  ndraws = 20
  sample1 = rand(x,20)

  plot(x=1:1:ndraws,y=sample1)

  # Can also truncate
  x_truncated = Truncated(x,-2.3 , 7.3)
  garbage = 0.0
end

# Then for a funkier distribution
function beta_test!(garbage)
  # First, set the relevant parameters
  α_x = 2.3
  β_x = 4.65

  # Next, define the variable
  x = Beta(α_x,β_x)

  # You can ask it for the pdf

  y_support = pdf(x,.2)
  y_no_support = pdf(x,1.2) # beta is always on [0,1]

  # You could even evaluate it at a bunch of points and plot it
  N = 1000
  ggrid = 0.0:1.0/(N-1):1

  pdf_vec = pdf(x,ggrid)

  plot(x=ggrid,y=pdf_vec,Geom.line)

  # You can do the same for the CDF
  cdf_vec = cdf(x,ggrid)

  plot(x=ggrid,y=cdf_vec,Geom.line)

  # you can get the relevant parameters with the params() command
  x_params = params(x)

  # You can also compute a bunch of moments
  x_mean = mean(x)
  x_median = median(x)
  x_mode = mode(x) # returns first mode if there are multiple
  x_skewness = skewness(x)
  x_kurtosis = kurtosis(x) # excess kurtosis over normal
  x_5perc = quantile(x,.05)
  x_95perc = quantile(x,.95)

  # You can even get the entropy!
  x_entropy = entropy(x)

  # And the log-likelihood of a bunch of draws
  x_sample = [.9,.4,.3]
  sampleLL = loglikelihood(x,x_sample)

  # You can also take random draws from it
  ndraws = 20
  sample1 = rand(x,20)

  plot(x=1:1:ndraws,y=sample1)

  garbage = 0.0
end

# it supports discrete distributions too
function binomial_test!(garbage)
  # First, set the relevant parameters
  n = 20
  p = .2

  # Next, define the variable
  x = Binomial(n,p)

  # You can ask it for the pdf

  y_support = pdf(x,2) # uses the pmf
  y_no_support = pdf(x,1.2) # Binomial only has support on integers

  # You could even evaluate it at a bunch of points and plot it
  N = n
  ggrid = 0:1:N

  pdf_vec = pdf(x,ggrid)

  plot(x=ggrid,y=pdf_vec,Geom.line)

  # Notice that the pmf must sum to one
  this_is_one = sum(pdf_vec)


  # You can do the same for the CDF
  cdf_vec = cdf(x,ggrid)

  plot(x=ggrid,y=cdf_vec,Geom.line)

  # you can get the relevant parameters with the params() command
  x_params = params(x)

  # You can also compute a bunch of moments
  x_mean = mean(x)
  x_median = median(x)
  x_skewness = skewness(x)
  x_kurtosis = kurtosis(x) # excess kurtosis over normal
  x_5perc = quantile(x,.05)
  x_95perc = quantile(x,.95)

  # You can even get the entropy!
  x_entropy = entropy(x)

  # And the log-likelihood of a bunch of draws
  x_sample = [1,12,7]
  sampleLL = loglikelihood(x,x_sample)

  # You can also take random draws from it
  ndraws = 20
  sample1 = rand(x,20)

  plot(x=1:1:ndraws,y=sample1)

  garbage = 0.0
end

# And multivariate
function multivariate_test!(garbage)
  # First, set the relevant parameters
  mean_x = [2.3,-1.2]
  ssigma_x = [4.65 -.2; -.2 1.2]

  # Next, define the variable
  x = MvNormal(mean_x,ssigma_x)

  # You can ask it for the pdf

  y = pdf(x,[8.4,1.1])

  x_params = params(x)

  # You can also compute a bunch of moments
  x_mean = mean(x)
  x_mode = mode(x) # returns first mode if there are multiple

  # You can even get the entropy!
  x_entropy = entropy(x)

  x_cor = cor(x)

  # You can also take random draws from it
  ndraws = 20
  sample1 = rand(x,20)

  garbage = 0.0
end

# You can even do simple maximum likelihood
some_data = [.2 .023 .98 .34 .25 .76 .223]

d = fit_mle(Normal,some_data)
# now take a look at the coefficients
(mle_normal_mean,mle_normal_std) = params(d)

d = fit_mle(Gamma,some_data)
# now take a look at the coefficients
(mle_α,mle_β) = params(d)




# Now we try a few different methods to compute numerical integrals.
# We'll start with a normal pdf, which should integrate to one
x_rv = Normal(1.25, .4)
our_func(x) = x*pdf(x_rv,x)
# we seek to integrate this function from [0,2.5]. It should be pretty darn close
# to 1.25
xHigh = 2.5
xLow = 0.0

# start with the box method
N = 7 # number of intervals + 1; try 3, 5, 7
xgrid = xLow:(xHigh - xLow)/(N-1): xHigh

rect_method = 0.0
trap_method = 0.0
for i in 2:N
  rect_method += (xgrid[i] - xgrid[i-1])*our_func((xgrid[i] + xgrid[i-1])/2.0)
  trap_method +=(xgrid[i] - xgrid[i-1])*(our_func(xgrid[i]) + our_func(xgrid[i-1]))/2.0
end
showall([rect_method, trap_method])



# Now, try it with Newton-Cotes
# Try it for 5
N = 5
xgrid = xLow:(xHigh - xLow)/(N-1): xHigh
h = xgrid[2] - xgrid[1]
# Now we evaluate it at each point
yvals = our_func.(xgrid) # notice the . after the function. Implies element-by-element operation

NC_method5 = (2/45)*h*( 7*yvals[1] + 32*yvals[2] + 12*yvals[3] + 32*yvals[4] + 7*yvals[5] )
showall(NC_method5)

N = 7 # increase it to 7
xgrid = xLow:(xHigh - xLow)/(N-1): xHigh
h = xgrid[2] - xgrid[1]
# Now we evaluate it at each point
yvals = our_func.(xgrid) # notice the . after the function. Implies element-by-element operation

NC_method7 = (1/140)*h*( 41*yvals[1] + 216*yvals[2] + 27*yvals[3] + 272*yvals[4] +
  27*yvals[5] + 216*yvals[6] + 41*yvals[7] )
showall(NC_method7)

# Finally up to 11
N = 11
xgrid = xLow:(xHigh - xLow)/(N-1): xHigh
h = xgrid[2] - xgrid[1]
# Now we evaluate it at each point
yvals = our_func.(xgrid) # notice the . after the function. Implies element-by-element operation

NC_method11 = (5/299376)*h*( 16067*(yvals[1]+yvals[11]) +
  106300*(yvals[2]+yvals[10]) -48525*(yvals[3]+yvals[9]) +
  272400*(yvals[4] + yvals[8]) - 260550*(yvals[5] + yvals[7]) +
  427368*yvals[6]  )
showall(NC_method11)

# subdivision method (often an improvement over higher N)
N = 8 # break it into 7 sections
# then subdivide them into 3, 2, and 2
xgrid = xLow:(xHigh - xLow)/(N-1): xHigh
h = xgrid[2] - xgrid[1]
# Now we evaluate it at each point
yvals = our_func.(xgrid) # notice the . after the function. Implies element-by-element operation


NC_subdivide = h*( (3/8)*yvals[1] + (9/8)*yvals[2] + (9/8)*yvals[3] +
  (17/24)*yvals[4] + (4/3)*yvals[5] + (2/3)*yvals[6] + (4/3)*yvals[7] + (1/3)*yvals[8] )
showall(NC_subdivide)


# Now we do it with the Gaussian quadrature method
# First, we translate the function
trans_func(x) = our_func((xHigh-xLow)/2*x + (xHigh+xLow)/2)

N = 4 # choose 4
xgrid = [-sqrt(525+70*sqrt(30))/35, -sqrt(525-70*sqrt(30))/35,
  sqrt(525-70*sqrt(30))/35, sqrt(525+70*sqrt(30))/35 ]

weights = [(18-sqrt(30))/36, (18+sqrt(30))/36,
    (18+sqrt(30))/36, (18-sqrt(30))/36  ]

gauss_method = (xHigh-xLow)/2*sum(weights.*trans_func.(xgrid))
showall(gauss_method)
# still a bit off, but better than NC_method5


# Could subdivide it into 2 sections to get a more accurate answer
xHighAlt = (xHigh+xLow)/2
xLowAlt = (xHigh+xLow)/2

trans_func1(x) = our_func((xHighAlt-xLow)/2*x + (xHighAlt+xLow)/2)
trans_func2(x) = our_func((xHigh-xLowAlt)/2*x + (xHigh+xLowAlt)/2)

gauss_method_alt = (xHighAlt-xLow)/2*sum(weights.*trans_func1.(xgrid)) +
  (xHigh-xLowAlt)/2*sum(weights.*trans_func2.(xgrid))
showall(gauss_method_alt)
# a lot closer without much more work!


# Can use the FastGaussQuadrature package to derive the nodes and weights for larger
# expressions
xgrid500, weights500 = gausslegendre( 500 )

gauss_method500 = (xHigh-xLow)/2*sum(weights500.*trans_func.(xgrid500))
showall(gauss_method500)
# not necessarily better than subdivision tricks






# Now we give it a shot with Monte Carlo methods
Ndraw = 50000 # increase Ndraw to improve accuracy

our_sample = rand(x_rv,Ndraw)

MC_method = sum(our_sample)/Ndraw
showall(MC_method)

# srand(1234) # can set the seed so it's the same each time




# Markov Chains
# Suppose we have a simple transition matrix
P = [.6 .39 .01; .1 .85 .05 ; .02 .03 .95]

# Computing the stationary distribution: Method 1
K = 1000
ψ0 = [.25 .5 .25]' # an arbitrary distribution
ψstar1 = vec( (P'^K)*ψ0 ) # matrix operations always give arrays

# Computing the stationary distribution: Method 2
(vals, vecs) = eig(eye(3) - P')

for i in 1:3
  if abs(vals[i]) < 1e-6
    global ψstar2 = vecs[:,i]/sum(vecs[:,i])
    break
  end
end



# Utility max example
# We'll keep the matrix P from above

# Y-values
Y = [.8 1 1.3]'
u = log(Y) # hand-to-mouth
β = .95

init_states = eye(3)

U = zeros(3,1)

for i in 1:3
  ttemp = init_states[:,i]'/(eye(3) - β*P)*u
  # returns a 1-dimensional array...can't technically assign a whole array to an element of another array
  # store it in a temp vector and pull out the first (only) element
  U[i] = ttemp[1]
end


# Tauchenize an AR(1) process
ρz = .999
σϵ = .025
σz = σϵ/√(1-ρz^2)
μz = .08

N = 7
m = 3.0 # stdevs from mean we want our grid to go

zLB = μz - m*σz
zUB = μz + m*σz

zgrid = zLB:(zUB-zLB)/(N-1): zUB

w = zgrid[2] - zgrid[1] # uniform distance between points

std_normal = Normal(0,1)
F(x) = cdf(std_normal,x)

trans_mat = zeros(N,N)

# Compute the transition probabilities
for i in 1:N
  trans_mat[i,1] = F( (zgrid[1] + w/2 - (1-ρz)*μz - ρz*zgrid[i])/σϵ )
  for j in 2:N-1
    trans_mat[i,j] = F( (zgrid[j] + w/2 - (1-ρz)*μz - ρz*zgrid[i])/σϵ ) -
      F( (zgrid[j] - w/2 - (1-ρz)*μz - ρz*zgrid[i])/σϵ )
  end
  trans_mat[i,N] = 1-sum(trans_mat[i,1:N-1])
end
# show poor performance for high ρz (.999)

# Can use transition matrix to do anything you like e.g. simulate, compute
# ergodic distributions, solve a model, etc.
